/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "main_header.h"
#include "engine_game.h"
#include "engine_physics.h"
#include "engine_graphics.h"
#include <rgTime.h>
#include <rgNetwork.h>
#include <rg3DRom.h>
#include "things/things.h"
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <string>


ConsoleMessage gbl_cmesg;

GameEngine::GameEngine()
{
	mdata = new MainData( this);
	m_shutdown = false;
	
	// Init network engine
	m_use_console_server = false;
	m_console = new GameConsole( this);

	// Init timing engine
	m_game_clock_scalar = 1.0;
	m_game_clock_hz = 100;

	// Init physics
	phy_engine = new PhysicsEngine( this);

	// Init graphics
	gfx_engine = new GraphicsEngine( this);

	// Init thing manager
	m_thing_manager = 0;

	m_batch_limit = 32;

	// Initialize console commands
	add_console_command( "add_bounding_thing", ccmd_add_bounding_thing);
	add_console_command( "add_camera", ccmd_add_camera);
	add_console_command( "batch", ccmd_batch);
	add_console_command( "bind", ccmd_bind);
	add_console_command( "bind_front", ccmd_bind_front);
	add_console_command( "bindings", ccmd_bindings);
	add_console_command( "build_sector_map", ccmd_build_sector_map);
	add_console_command( "clear_color", ccmd_clear_color);
	add_console_command( "clear_sector_map", ccmd_clear_sector_map);
	add_console_command( "color_material", ccmd_color_material);
	add_console_command( "compile_model_alpha_faces",
		ccmd_compile_model_alpha_faces);
	add_console_command( "cross", ccmd_cross);
	add_console_command( "cull_face", ccmd_cullface);
	add_console_command( "depth_test", ccmd_depth_test);
	add_console_command( "do", ccmd_do);
	add_console_command( "dot", ccmd_dot);
	add_console_command( "draw_sectors", ccmd_draw_sectors);
	add_console_command( "dump_all_things", ccmd_dump_all_things);
	add_console_command( "dump_camera", ccmd_dump_camera);
	add_console_command( "dump_model", ccmd_dump_model);
	add_console_command( "dump_sector_map", ccmd_dump_sector_map);
	add_console_command( "dump_thing", ccmd_dump_thing);
	add_console_command( "echo", ccmd_echo);
	add_console_command( "game_clock", ccmd_game_clock);
	add_console_command( "game_clock_scalar", ccmd_game_clock_scalar);
	add_console_command( "gfx_engine", ccmd_gfx_engine);
	add_console_command( "gravity", ccmd_gravity);
	add_console_command( "kill", ccmd_kill);
	add_console_command( "list_cameras", ccmd_list_cameras);
	add_console_command( "list_commands", ccmd_list_commands);
	add_console_command( "list_models", ccmd_list_models);
	add_console_command( "list_spawnable", ccmd_list_spawnable);
	add_console_command( "load_model", ccmd_load_model);
	add_console_command( "load_model_textures", ccmd_load_model_textures);
	add_console_command( "load_module", ccmd_load_module);
	add_console_command( "load_texture", ccmd_load_texture);
	add_console_command( "normalize", ccmd_normalize);
	add_console_command( "perspective", ccmd_perspective);
	add_console_command( "ping", ccmd_ping);
	add_console_command( "pwd", ccmd_pwd);
	add_console_command( "quit", ccmd_quit);
	add_console_command( "remove_camera", ccmd_remove_camera);
	add_console_command( "remove_model", ccmd_remove_model);
	add_console_command( "remove_thing", ccmd_remove_thing);
	add_console_command( "set_alt_texture", ccmd_set_alt_texture);
	add_console_command( "set_camera_position", ccmd_set_camera_position);
	add_console_command( "set_current_camera", ccmd_set_current_camera);
	add_console_command( "set_game_clock_hz", ccmd_set_game_clock_hz);
	add_console_command( "set_root_renderer", ccmd_set_root_renderer);
	add_console_command( "set_screen", ccmd_set_screen);
	add_console_command( "set_server_port_range",
		ccmd_set_server_port_range);
	add_console_command( "set_texture_color_key",
		ccmd_set_texture_color_key);
	add_console_command( "spawn", ccmd_spawn);
	add_console_command( "thing_list_check", ccmd_thing_list_check);
	add_console_command( "thing_manager", ccmd_thing_manager);
	add_console_command( "unbind", ccmd_unbind);
	add_console_command( "use_console_server", ccmd_use_console_server);
	add_console_command( "wireframe", ccmd_wireframe);

	m_engine_started = false;
	rcout( "Loading bootupt file \"" BOOTUP_FILE_NAME "\"...\n");
	const int err = process_batch_file( BOOTUP_FILE_NAME, true);
	// FIXME: Don't use magic error numbers.
	switch( err) {
	case 0: rcout( "bootup file loaded\n"); break;
	case -1: rcout( "no bootup file present.\n"); break;
	default:
		std::cerr << "Error on line" << err <<
			": Failed to load boot file \""
			BOOTUP_FILE_NAME "\".\n";
		m_shutdown = true;
		return;
	}

	if( m_use_console_server) {
		if( m_console->start_server_thread())
			std::cerr << "Could not start server thread\n";
	}

	// Finialize timing engine
	m_game_clock_tick = 1.0 / double( m_game_clock_hz);
	m_game_timer = new rgTimer( m_game_clock_tick);
	m_game_clock = 0;
}

GameEngine::~GameEngine()
{
	clear();
	if( mdata) delete mdata;
}

void GameEngine::clear()
{
	if( m_game_timer) { delete m_game_timer; m_game_timer = 0;}
	if( m_thing_manager) { delete m_thing_manager; m_thing_manager = 0;}
	if( gfx_engine) { delete gfx_engine; gfx_engine = 0;}
	if( phy_engine) { delete phy_engine; phy_engine = 0;}
	if( m_console) { delete m_console; m_console = 0;}
}

int GameEngine::startup()
{
	gfx_engine->startup();
	create_thing_manager();
	m_engine_started = true;
	rcout( "loading startup file...\n");
	const int err = process_batch_file( STARTUP_FILE_NAME, true);
	// FIXME: Don't use magic error numbers.
	switch( err) {
	case 0: rcout( "startup file loaded\n"); break;
	case -1: rcout( "startup: no startup file present."); break;
	default:
		rcout( "startup: "
			"Error on line " + gm_to_string( err) +
			": while loading startup file \"" STARTUP_FILE_NAME
			"\".\n");
		return -2;
	}
	return 0;
}

// Returns: "true" is returned if given (key, function) pair was added,
//	"false" is returned otherwise.
bool GameEngine::add_console_command( const std::string &key,
	int (*fn)( const std::vector <std::string> &, GameEngine &))
{
#define TMP_PROTO int (*)( const std::vector <std::string> &, \
	GameEngine &)
	typedef std::pair <const std::string, TMP_PROTO> SCPair;
	return (m_commands.insert( SCPair( key, fn))).second;
#undef TMP_PROTO
}

gmInputAction *GameEngine::binding( const std::string &key)
{
	std::map <const std::string, gmInputAction> :: iterator iter =
		m_key_binding.find( key);
	if( iter == m_key_binding.end()) return 0;
	return &(iter->second);
}

int GameEngine::add_binding( const std::string &key,
	const std::vector <std::string> &args,
	const std::string &up_down,
	bool to_front)
{
	gmInputAction *ia = binding( key);
	if( !ia) {
		typedef std::pair <const std::string, gmInputAction> SAPair;
		m_key_binding.insert( SAPair( key, gmInputAction()));
		ia = binding( key);
		if( !ia) return -1;
	}
	std::string tmp;
	for( int i = 0; i < int( args.size()); i++) tmp += args[i] + " ";
	std::vector <std::string> *p = &(ia->m_down);
	if( "up" == up_down) p = &(ia->m_up);
	if( to_front) p->insert( p->begin(), tmp);
	else p->push_back( tmp);
	return 0;
}

int GameEngine::unbind( const std::string &key)
{
	std::map <const std::string, gmInputAction> :: iterator iter =
		m_key_binding.find( key);
	if( iter == m_key_binding.end()) return -1;
	m_key_binding.erase( iter);
	return 0;
}

std::string GameEngine::dump_bound_keys()
{
	std::map <const std::string, gmInputAction> :: const_iterator iter =
		m_key_binding.begin();
	std::string tmp( "Key Bindings:\n");
	for( ; iter != m_key_binding.end(); iter++)
		tmp += dump_binding( iter->first);
	tmp += "Total bound keys: " + gm_to_string( int(
		m_key_binding.size())) + "\n";
	return tmp;
}

std::string GameEngine::dump_binding( const std::string &key)
{
	gmInputAction *ia = binding( key);
	if( !ia) return std::string( "No bindings for key \"" + key + "\"\n");
	std::string tmp( "Key \"" + key + "\":\n");

	{	tmp += "\t[up] bindings:\n";
		for( int i = 0; i < int( ia->m_up.size()); i++)
			tmp += "\t\t" + ia->m_up[i] + "\n";
	}

	{	tmp += "\t[down] bindings:\n";
		for( int i = 0; i < int( ia->m_down.size()); i++)
			tmp += "\t\t" + ia->m_down[i] + "\n";
	}
	return tmp;
}

int GameEngine::exec_key( const std::string &key, const std::string &up_down)
{
	gmInputAction *ia = binding( key);
	if( !ia) return -1;
	if( "up" == up_down) {
		for( int i = 0; i < int( ia->m_up.size()); i++)
			process_message( ia->m_up[i]);

	} else {
		for( int i = 0; i < int( ia->m_down.size()); i++)
			process_message( ia->m_down[i]);
	}
	return 0;
}

//int GameEngine::load_arena( const char *fname)
//{
//	if( !gfx_engine) return -1;
//	return gfx_engine->load_arena( fname);
//}

int GameEngine::process()
{
	check_for_console_messages();

	long long clk = game_clocks_pending();
	int clocks_processed = 0;
	while( clk--) {
		m_thing_manager->process();

		m_game_clock++;
//		if( !(game_clock() % m_game_clock_hz)) {
//			std::cerr << "game second passed. " << game_clock()
//				<< "\n";
//		}
	
		// FIXME: Check console messages frequently enough to remain
		//	responsive.
		//	However, more events must be checked to keep the game
		//	responsive, implement a better system.
		if( (clocks_processed++) > 100) {
			check_for_console_messages();
			clocks_processed = 0;
		}
	}

	gfx_engine->process();

	if( m_shutdown) return -1;
	return 0;
}

long long GameEngine::game_clocks_pending()
{
	// FIXME: Game clock scaling does not work properly if a scaling value
	//	greater than 1 is given. Also, small scaling values pause the
	//	game (since clocks pending is integral). Therefore think of
	//	a better implementation.
	if( m_game_timer->time_unit_passed()) {
		return (long long)(
			(m_game_timer->elapsed_mili_sec() *
				double( m_game_clock_hz * m_game_clock_scalar))
				/ 1000.0);
	}
	else return 0;
}

int GameEngine::load_startup_file( const char *fname)
{
	return 0;
}

int GameEngine::create_thing_manager()
{
	if( m_thing_manager) return 0;
	if( !(m_thing_manager = new ThingManager( this))) return -1;
	return 0;
}

void GameEngine::check_for_console_messages()
{
	if( gbl_cmesg.console_lock) return;

	std::vector <std::string> msg;
	std::string tmp;
	if( m_pending_console_cmd.length() > 0)
		tmp = m_pending_console_cmd;
	for( int i = 0; i < gbl_cmesg.length; i++){
		if( '\0' == gbl_cmesg.msg[i]) {
			if( tmp.length() > 0) {
				msg.push_back( tmp);
				tmp.clear();
			}
		} else tmp += gbl_cmesg.msg[i];
	}
	if( tmp.length() > 0) m_pending_console_cmd = tmp;
	else m_pending_console_cmd.clear();

	while( !msg.empty()) {
		//std::cerr << "ConMsg: \"" << msg[0] << "\"\n";
		process_message( msg[0]);
		msg.erase( msg.begin());
	}

	gbl_cmesg.console_lock = true;
	gbl_cmesg.game_lock = false;
}

GameConsole::GameConsole( GameEngine *p_ge)
{
	ge = p_ge;
	m_sock = new rgSocket();
	m_first_port = 9000;
	m_last_port = 16000;
	m_server_started = false;
}

GameConsole::~GameConsole()
{
	if( m_server_started) {
/*
		std::cout << "Shutting down console server\n";
		std::cout.flush();

		// FIXME: Must shutdown console server thread, otherwise the
		// program will segfault when "this"
		// (GameConsole::~GameConsole) returns because its memory will
		// still be accessed by the console server thread.
		// pthread_cancel and pthread_join are not working.
		std::cerr << "Waiting for server thread to terminate...";
		int err = pthread_cancel( m_server_thread);
		if( err) {
			std::cerr << "Error " << err <<
				" while canceling server thread.\n";
		}

		if( (err = pthread_join( m_server_thread, 0))) {
			std::cerr << "Error " << err <<
				" while joining server thread.\n";

		} else std::cerr << "done\n";
*/
	}

	// FIXME: Remove the below "exit".
	exit(0);
	//if( m_sock) delete m_sock;
}

int GameConsole::start_server_thread()
{
	int result = pthread_create( &m_server_thread, NULL,
		console_server_thread, this);
	if( result) return -1;
	m_server_started = true;
	return 0;
}

void ConsoleMessage::print_message( std::ostream &o)
{
	std::string tmp;
	for( int i = 0; i < length; i++) {
		if(	((msg[i] >= 'A') && (msg[i] <= 'z')) ||
			((msg[i] >= ' ') && (msg[i] <= '@')) ||
			(msg[i] == '\t') ||
			(msg[i] == '\n')) tmp += msg[i];
		else tmp += '.';
	}
	o << tmp;
}

int hlp_get_message( GameConsole *gc)
{
	while( gbl_cmesg.game_lock) sleep( 1);
	gbl_cmesg.length = gc->m_sock->receive( gbl_cmesg.msg,
		MAX_CONSOLE_MESSAGE_LENGTH - 1);
	if( -1 == gbl_cmesg.length) {
		std::cerr << "Error \"" << strerror( errno) <<
			"\" while receiving console message\n";
		sleep( 1);	// XXX: Required to prevent the console
				// messaging thread from hogging the cpu.
		return errno;
	}

	gbl_cmesg.game_lock = true;
	gbl_cmesg.console_lock = false;
	return 0;
}

void *console_server_thread( void *ptr)
{
	GameConsole *gc = (GameConsole *)ptr;
	int start_port = gc->m_first_port;
	while(	gc->m_sock->open_listen_port( start_port) &&
		(start_port < gc->m_last_port)) start_port++;
	if( start_port >= gc->m_last_port) {
		std::cerr << "Could not find a free port number for server\n";
		return 0;
	}
	std::cout << "Started console server on port " << start_port << "\n";
	std::cout.flush();

	if( -1 == gc->m_sock->accept()) {
		std::cerr << "Error: \"" << strerror( errno) <<
			"\" while accepting console server connections\n";
		return 0;
	}
	std::cout << "Remote console connected\n";
	std::cout << "Console located at " << gc->m_sock->remote_ip() << "\n";
	std::cout.flush();

	while( !(hlp_get_message( gc)));
	std::cerr << "Remote console terminated\n";
	return 0;
}

