/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#ifndef ENGINE_GRAPHICS_H
#define ENGINE_GRAPHICS_H

#include <list>
#include <map>
#include <string>
#include <rgStandard.h>
#include <rgOpenGL.h>

class sk3DRomModel;
class ModelManager;
class gmThing;
class ThingControlBlock;
class TextureHandle;

class gmCamera : public rgGLCamera
{
public:
	void set_owner( gmThing *p) { m_owner = p;}
	gmThing *owner() const { return m_owner;}

	gmCamera() { m_owner = 0;}
	gmCamera( rgVector p, rgVector p_look_at, rgVector up) :
		rgGLCamera( p, p_look_at, up) {
		m_owner = 0;
	}

protected:
	gmThing *m_owner;
};

class GraphicsEngine
{
public:

	bool m_enabled;

	int m_draw_sector_level;
	int m_sector_line_width;
	bool m_draw_sectors;

	GraphicsEngine( GameEngine *ge);
	~GraphicsEngine();

	void clear();
	void process();

	bool set_current_camera( const std::string &key);
	gmCamera *current_camera() {
		std::map <const std::string, gmCamera *> :: const_iterator
			iter = m_camera.find( m_current_camera);
		if( iter == m_camera.end()) return 0;
		return (iter->second);
	}
	int add_camera( const std::string &key);
	int add_camera( const std::string &key, rgVector pos,
		rgVector p_look_at);
	int add_camera( const std::string &key, rgVector pos,
		rgVector p_look_at, rgVector up);

	int remove_camera( const std::string &key);
	int remove_camera( std::map <const std::string, gmCamera *> ::
		iterator &iter);

	std::string list_cameras() const;

	gmCamera *get_camera( const std::string &key);

	std::string dump_camera( const std::string &key) const;
	std::string dump_camera( std::map <const std::string, gmCamera *> ::
		const_iterator &iter) const;

	ModelManager *model_man() { return m_model_man;}

	void reset_all_cameras();

	bool set_root_renderer( const std::string &key);

	TextureHandle *add_texture( const std::string &key);
//	int load_texture( const std::string &key, const std::string &fname,
//		const std::vector <std::string> &args);
	int delete_texture( const std::string &key);
	TextureHandle *get_texture( const std::string &key);

	void startup();

private:
	GameEngine *game_engine;
	ModelManager *m_model_man;

	// root_renderer
	// This is the thing that manages all rendering for the graphics	
	// engine. Every time the graphics engine's "process" function is
	// called, root_renderer's "process" function is also called to
	// perform the rendering. Therefore, it is best for root_renderer to
	// be a thing that is in the "sleeping list" so that its "process"
	// function is only called when graphics processing is needed.
	ThingControlBlock *root_renderer;

	std::string m_current_camera;
	std::map <const std::string, gmCamera *> m_camera;

	std::map <const std::string, TextureHandle *> m_texture;

	void init_gl_state();
	void set_projection();

	int delete_texture( std::map <const std::string, TextureHandle *> ::
		iterator &iter);
};

class TextureHandle
{
public:
	TextureHandle();
	~TextureHandle();

	rgGLTexture *m_tex;

	int reference_count() const { return m_ref_count;}
	void inc_reference() { m_ref_count++;}
	void dec_reference() { if( m_ref_count > 0) m_ref_count--;}

private:
	int m_ref_count;

	TextureHandle( const TextureHandle &);
	TextureHandle *operator ==( const TextureHandle &);
};


#define MAX_MODEL_FILENAME_LENGTH 280
class ModelHandle
{
public:
	char fname[ MAX_MODEL_FILENAME_LENGTH];
	sk3DRomModel *model;

	ModelHandle( const char *f);
	~ModelHandle();

	int load();
	int reference_count() const { return m_ref_count;}
	void inc_reference() { m_ref_count++;}
	void dec_reference() { if( m_ref_count > 0) m_ref_count--;}

	void build_bone_name_map();
	void build_frame_name_map();
	void build_texture_name_map();
	void build_group_name_map();
	void build_sequence_name_map();

	int frame_id( const std::string &key) const {
		std::map <const std::string, int> :: const_iterator iter =
			m_frame.find( key);
		if( iter == m_frame.end()) return -1;
		return (iter->second);
	}

	int bone_id( const std::string &key) const {
		std::map <const std::string, int> :: const_iterator iter =
			m_bone.find( key);
		if( iter == m_bone.end()) return -1;
		return (iter->second);
	}

	int group_id( const std::string &key) const {
		std::map <const std::string, int> :: const_iterator iter =
			m_group.find( key);
		if( iter == m_group.end()) return -1;
		return (iter->second);
	}

	int sequence_id( const std::string &key) const {
		std::map <const std::string, int> :: const_iterator iter =
			m_sequence.find( key);
		if( iter == m_sequence.end()) return -1;
		return (iter->second);
	}

	int texture_id( const std::string &key) const {
		std::map <const std::string, int> :: const_iterator iter =
			m_texture.find( key);
		if( iter == m_texture.end()) return -1;
		return (iter->second);
	}

	int set_alternate_texture( const std::string &des,
		const std::string &src, GraphicsEngine *gfx);

	std::string dump();
	std::string dump_bones();
	std::string dump_frames();
	std::string dump_textures();
	std::string dump_groups();
	std::string dump_sequences();

protected:
	std::map <const std::string, int> m_bone;
	std::map <const std::string, int> m_frame;
	std::map <const std::string, int> m_texture;
	std::map <const std::string, int> m_group;
	std::map <const std::string, int> m_sequence;

	std::map <int, const std::string> m_alt_texture;

private:
	int m_ref_count;

	ModelHandle( const ModelHandle &);
	ModelHandle *operator ==( const ModelHandle &);
};

#define MODEL_MANAGER_ERR_NO_MODEL	-1
#define MODEL_MANAGER_ERR_REFERENCE	-2
class ModelManager
{
public:

	ModelHandle *add( const std::string &key, const std::string &fname);
	ModelHandle *get( const std::string &key);
	std::map <const std::string, ModelHandle *> :: const_iterator
		get_begin_iterator();
	std::map <const std::string, ModelHandle *> :: const_iterator
		get_end_iterator();

	int remove( const std::string &key);

	ModelHandle *has_model( const sk3DRomModel *p_model);
	
	int model_count() const { return int( m_model.size());}

	~ModelManager();
	void clear();

protected:
	std::map <const std::string, ModelHandle *> m_model;

};

#endif
