/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#ifndef ENGINE_PHYSICS_H
#define ENGINE_PHYSICS_H

#include <rgStandard.h>
#include <rgPhysics.h>
#include <map>
#include <string>

class gmModelThing;
class gmPhyModel;
class GameEngine;
class sk3DRomSkelFrame;
class sk3DRomPhySkel;
class sk3DRomPhyBone;

class PhyBoundingPoint : public rgVector
{
public:
	float radius;

	rgVector m_origin;

	PhyBoundingPoint() { radius = 0;}
	PhyBoundingPoint( const rgVector p, const float r) : rgVector( p) {
		radius = r;
	}

	PhyBoundingPoint( const rgVector p) : rgVector( p) {
		radius = 0.0f;
	}

	PhyBoundingPoint &operator =( const rgVector p) {
		x = p.x; y = p.y; z = p.z;
		return *this;
	}
};

class PhyCollisionPrisim : public rgCollisionPrisim
{
public:
	float m_friction;

	PhyCollisionPrisim() {
		m_friction = 0.0025f;
	}

	PhyCollisionPrisim &create( float x_scale, float y_scale,
		float z_scale) {
		rgCollisionPrisim::create( x_scale, y_scale, z_scale);
		return *this;
	}

	PhyCollisionPrisim &create( const std::vector <rgVector> &v,
		const rgMatrix &m) {
		rgCollisionPrisim::create( v, m);
		return *this;
	}
	
};

class PhyCollisionFragment
{
public:
	int pid;	// Index number of bounding point for this fragment
	PhyCollisionPrisim *prisim;	// Colliding prisim
	rgVector src_point;
	rgVector des_point;
	rgVector surface_point;
	rgVector surface_normal;
	rgVector collision_point;
	rgVector intrusion_vector;
	float radius;

	PhyCollisionFragment(){}
	PhyCollisionFragment( int p_pid, PhyCollisionPrisim *p) {
		pid = p_pid;
		prisim = p;
	}

	std::string dump() const;
};

#define PHY_MAX_SUB_SECTORS	4
class PhySector
{
public:
	rgVector m_v[4];
	float m_width;
	float m_depth;
	std::vector <PhyCollisionPrisim *> m_prisim;

	PhySector( rgVector a, rgVector b, rgVector c, rgVector d) {
		m_sec[0] = m_sec[1] = m_sec[2] = m_sec[3] = 0;
		m_v[0] = a;
		m_v[1] = b;
		m_v[2] = c;
		m_v[3] = d;
		m_width = (a - b).length();
		m_depth = (a - d).length();
	}
	~PhySector() {
		for( int i = 0; i < PHY_MAX_SUB_SECTORS; i++)
			if( m_sec[i]) delete m_sec[i];
	}

	// B|A
	// -+-
	// C|D
	PhySector *sec_a() { return m_sec[0];}
	PhySector *sec_b() { return m_sec[1];}
	PhySector *sec_c() { return m_sec[2];}
	PhySector *sec_d() { return m_sec[3];}
	const PhySector *const_sec_a() const { return m_sec[0];}
	const PhySector *const_sec_b() const { return m_sec[1];}
	const PhySector *const_sec_c() const { return m_sec[2];}
	const PhySector *const_sec_d() const { return m_sec[3];}


	rgVector v_a() const { return m_v[0];}
	rgVector v_b() const { return m_v[1];}
	rgVector v_c() const { return m_v[2];}
	rgVector v_d() const { return m_v[3];}

	const PhySector *containing_quadrant( const rgVector p) const;


	rgVector center() const { return (v_a() + v_c()) * 0.5f;}

	bool create_child_sector( int n);

	void draw( const int level) const;

	void sub_divide( const std::vector <PhyCollisionPrisim *> &p_prisim,
		const int level);

	void add_contained_prisims( const std::vector <PhyCollisionPrisim *>
		&p_prisim);

	std::string dump( const int p_depth);

	//PhyCollisionPrisim *point_collides( const PhyBoundingPoint p) const;
	bool point_collides( const PhyBoundingPoint p,
		PhyCollisionFragment &f) const;

protected:
	PhySector *m_sec[4];

private:
	PhySector( const PhySector &);
	PhySector &operator =( const PhySector &);
};

class PhySectorMap
{
public:

	PhySectorMap() { m_root = 0;}
	~PhySectorMap() { clear();}

	bool add_prisims( gmModelThing *m);
	bool build_map( rgVector a, rgVector b, rgVector c, rgVector d);

	PhySector *m_root;
	std::vector <PhyCollisionPrisim> m_prisim;

	void draw( const int level, const int p_line_width) const;
	void clear();

	std::string dump();

	//PhyCollisionPrisim *point_collides( const PhyBoundingPoint p) const;
	bool point_collides( const PhyBoundingPoint p,
		PhyCollisionFragment &f) const;

private:
	PhySectorMap( const PhySectorMap &);
	PhySectorMap &operator =( const PhySectorMap &);
};

class PhyCollisionContext
{
public:
	gmPhyModel *m_model;

	// Bounding points for the physical model.
	std::vector <PhyBoundingPoint> m_bpoint;

	PhyCollisionContext( gmPhyModel *pm) { m_model = pm;}
};

class PhyDynResult
{
public:
//	bool hit;

//	PhyDynResult() {
//		hit = false;
//	}

	// Contains index numbers of bounding points that collided. If
	// m_point_hit contains no points, then there was no collision.
	std::vector <PhyCollisionFragment> m_point_hit;

	// Returns the collision center of mass.
	rgVector collision_center( const PhyCollisionContext &cc) const;

	std::string dump() const;
};

class PhysicsEngine
{
public:
	rgVector m_gravity;
	PhySectorMap m_sector_map;

	PhysicsEngine( GameEngine *ge);
	~PhysicsEngine();

	void get_origin_points( PhyCollisionContext &cc) const;
	void get_bounding_points( PhyCollisionContext &cc) const;

	// Description: Generates bounding points for the given physical model
	//	based on the models current skeleton, specified by the models
	//	draw context (m_draw_context).
	void gen_bounding_points( PhyCollisionContext &cc) const;

	bool environment_collision( PhyDynResult &result,
		const PhyCollisionContext &cc) const;

	float quadratic_frame_scalar() const;
	float friction_frame_scalar() const { return quadratic_frame_scalar();}
	float gravity_frame_scalar() const { return quadratic_frame_scalar();}

	// Description: Takes a physical model in an initial state "src" and
	//	tries to set it to the "des" state. The result of the physics
	//	performed on the model are returned. "des" will be left
	//	untouched if it can be set without violating any physical
	//	restrictions. Otherwise, "des" will be modified so that it
	//	does comply with restrictions. If all else fails, "des" will
	//	be set equal to "src" (since "src" is always in a valid
	//	physical state).
	PhyDynResult check_model( gmPhyModel *pm, sk3DRomSkelFrame &src,
		sk3DRomSkelFrame &des);

	bool check_for_collision( const int bcount, const sk3DRomPhySkel *s,
		bool *bcol);

	int hit( const sk3DRomPhyBone &b) const;


protected:
	GameEngine *game_engine;


private:
	PhysicsEngine( const PhysicsEngine &);
	PhysicsEngine &operator =( const PhysicsEngine &);
};

/*
// The sector table stores 2x2x2 sectors (or 8 cubic feet). This is sufficient
// to store the entire ice surface including benches and penatly box areas.
#define SECTOR_TABLE_WIDTH 60
#define SECTOR_TABLE_HEIGHT 100
#define SECTOR_TABLE_RESOLUTION (SECTOR_TABLE_WIDTH * SECTOR_TABLE_HEIGHT)

class GameEngine;
class gmPhyModel;
class sk3DRomPhySkel;
class sk3DRomPhyBone;
class sk3DRomSkelFrame;

#define PHYCHUNK_FRIEND_BIT	0x1
class PhyChunk
{
public:
	sk3DRomPhyBone *bone;
	short prev;
	short next;

	PhyChunk() { prev = -1; next = -1; bone = 0;}
};

class PhyTicket
{
public:
	short chunk_base;
	short total_chunks;
	PhyTicket() { chunk_base = -1; total_chunks = -1;}
};

class PhysicsEngine
{
public:
	int x_offset;
	int y_offset;
	int z_offset;

	rgVector m_gravity;

	int m_sector_width;
	int m_sector_height;

	short m_sector_table[SECTOR_TABLE_RESOLUTION];
	PhyChunk *chunk_table;

	PhysicsEngine( GameEngine *ge);
	~PhysicsEngine();

	// "count" is the total number of chunks that exsist.
	void set_chunk_table( int count);

	void check_model( gmPhyModel *pm, sk3DRomSkelFrame &src,
		sk3DRomSkelFrame &des);
	//void animate_model( gmPhyModel *pm, int p_elapsed);

	int get_sector( const sk3DRomPhyBone &b) const;

protected:
	GameEngine *game_engine;

	std::map <gmPhyModel *, const PhyTicket> m_things;

	int m_chunk_table_size;

	void init_sector_table();

	//
	//	id - The bone's chunk table index number.
	// Returns: 0 if bone was "sectorized". -1 if bone could not be
	//	sectorized (due to collision).
	int sectorize( int id, const sk3DRomPhyBone &b);

	int sec_hit( int id, const sk3DRomPhyBone &b) const;

	bool check_for_collision( const int bcount, const int chunk_base,
		const sk3DRomPhySkel *s, bool *bcol);
};
*/
#endif
