/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "things.h"
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"
#include <fstream>

std::map <const std::string, int (*)( const gmCCArg &)> gmThing::m_ccmd;

int gmThing_set_pos( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 3) return GM_THING_ERR_PARAM;
	p.thing->m_cs.p = gm_str_to_vector( args[0], args[1], args[2]);
	return 0;
}

int gmThing_set_velocity( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	const std::string usage( "usage: set_velocity <mode> <x> <y> <z>\n"
			"NOTE: \"mode\" can be one of \"add\", \"sub\", "
				"\"mul\" or \"set\"\n"
			"Each mode refers to how the given velocity is "
				"combined with the\n"
			"thing's current velocity\n"
			"\"x\", \"y\" and \"z\" are floating point "
				"numbers defining vector (x, y, z).\n"
			"Given velocity is in meters per second.\n"
			"(It will be converted internally into meters per "
				"frame.\n");
	if( args.size() != 4) {
		p.ge->rcout( usage);
		return GM_THING_ERR_PARAM;
	}
	const rgVector v( gm_str_to_vector( args[1], args[2], args[3])
		* (1.0f / double( p.ge->game_clock_hz())));
	if( "add" == args[0]) p.thing->m_velocity += v;
	else if( "sub" == args[0]) p.thing->m_velocity -= v;
	else if( "mul" == args[0]) {
		p.thing->m_velocity.x *= v.x;
		p.thing->m_velocity.y *= v.y;
		p.thing->m_velocity.z *= v.z;

	} else if( "set" == args[0]) p.thing->m_velocity = v;
	else {
		p.ge->rcout( "No such mode \"" + args[0] + "\"\n" + usage);
		return GM_THING_ERR_PARAM;
	}
	return 0;
}

int gmThing_set_gravity( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;
	p.thing->m_use_gravity = gm_set_state( args[0]);
	return 0;
}

int gmThing_move_to_proc( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;
	p.thing->move_to_proc_list( atoi( (args[0]).c_str()));
	p.ge->rcout( "Thing \"" + (p.thing->tcb())->m_key + "\" moved to "
		"process list " + gm_to_string( atoi( (args[0]).c_str())) +
		" ( " + gm_to_string( (p.thing->tcb())->m_proc) + ")\n");
	return 0;
}

void gmThing::move_to_proc_list( const int pid)
{
	((tcb())->manager())->move_thing_to_proc_list( tcb(), pid);
}

//int gmThing_set_csys( const gmCCArg &p)
//{
//	return 0;
//}

gmThing::gmThing()
{
	m_use_gravity = false;
	m_tcb = 0;

	// FIXME: Keep instance count of the class so that when all classes
	//      of this type are destroyed, the console command list is
	//      removed (though not doing so should waste few memory
	//      resources).
	// Prepare list of console commands for this class (only done once,
	// future instances will use what a previous instance created).
	if( !(m_ccmd.empty())) return;
	add_console_command( "set_pos", gmThing_set_pos, m_ccmd);
	add_console_command( "set_gravity", gmThing_set_gravity, m_ccmd);
	add_console_command( "move_to_proc", gmThing_move_to_proc, m_ccmd);
	add_console_command( "set_velocity", gmThing_set_velocity, m_ccmd);
	//add_console_command( "set_csys", gmThing_set_csys, m_ccmd);
}

PhysicsEngine *gmThing::phy_engine()
{
	return (((tcb())->manager())->ge())->phy_engine;
}

GameEngine *gmThing::ge()
{
	return ((tcb())->manager())->ge();
}

bool gmThing::add_console_command( const std::string &key,
	int (*fn)( const gmCCArg &),
	std::map <const std::string, int (*)(const gmCCArg &)> &p_ccmd)
{
	typedef std::pair <const std::string, int (*)( const gmCCArg &)>
		SAPair;
	return ((p_ccmd.insert( SAPair( key, fn))).second);
}

int gmThing::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.find( key);
	if( iter == m_ccmd.end()) return GM_THING_ERR_NO_CCMD;
	return (iter->second)( args);
}

std::string gmThing::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, int (*)( const gmCCArg &)> ::
		const_iterator iter = m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); iter++) {
		tmp += p_prefix + gmThing::type() + "::" + (iter->first) +
			"\n";
	}
	return tmp;
}

std::string gmThing::get_key()
{
	return (tcb())->m_key;
}
