/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#ifndef THINGS_H
#define THINGS_H

#include <rgStandard.h>
#include <rgOpenGL.h>
#include <vector>
#include <list>
#include <string>
#include <map>

class GameEngine;
class sk3DRomModel;
class sk3DRomDrawContext;
class sk3DRomPhySkel;
class ThingControlBlock;
class ThingManager;
class ModelHandle;
class ModelManager;
class gmThing;
class PhysicsEngine;
class PhyBoundingPoint;
class PhyCollisionContext;
class PhyCollisionFragment;

// gmRenderArgs
// Description: Defines how a "thing" should render itself. This class must
//	be modified to suit the desired application.
class gmRenderArgs
{
public:
	int lod_level;
	bool render_shadow;

	bool render_transparent;
	bool render_opaque;

	gmRenderArgs() { init();}

	gmRenderArgs( int lod, bool p_shadow) {
		init();
		lod_level = lod;
		render_shadow = p_shadow;
	}

private:
	void init() {
		lod_level = 0;
		render_shadow = false;
		render_opaque = true;
		render_transparent = false;
	}
};


// gmCCArg
// Description: Console Command Arguments.
class gmCCArg
{
public:
	gmThing *thing;
	GameEngine *ge;
	const std::vector <std::string> *args;

	gmCCArg( gmThing *t, GameEngine *g,
		const std::vector <std::string> *a) {
		thing = t;
		ge = g;
		args = a;
	}
};

// gmThing
// Description: The most basic "thing" that can exist in the game world. All
//	things derive from "gmThing".
#define GM_THING_ERR_GENERIC	-1
#define GM_THING_ERR_NO_CCMD	-2
#define GM_THING_ERR_PARAM	-3
class gmThing
{
public:
	gmThing();
	virtual ~gmThing(){}
	virtual rgCoordSystem cs() { return m_cs;}
	virtual void process( ThingManager *tm){}
	virtual void draw( gmRenderArgs *){}

	ThingControlBlock *tcb() { return m_tcb;}

	virtual void apply_velocity(){}
	virtual void apply_gravity(){}

	bool add_console_command( const std::string &key,
		int (*fn)( const gmCCArg &),
		std::map <const std::string, int (*)(const gmCCArg &)>
			&p_ccmd);

	virtual std::string type() const { return std::string( "gmThing");}
	virtual bool is_a( const std::string &s) const {
		if( s != gmThing::type()) return false;
		return true;
	}

	virtual int call_ccmd( const std::string &key, gmCCArg &args);
	virtual std::string list_ccmd( const std::string &p_prefix);

	void move_to_proc_list( const int pid);

	PhysicsEngine *phy_engine();
	GameEngine *ge();

	std::string get_key();

protected:
	rgCoordSystem m_cs;
	rgVector m_velocity;
	rgVector m_torque;

	bool m_use_gravity;

	// Map of supported console commands for the thing.
	static std::map <const std::string, int (*)( const gmCCArg &)> m_ccmd;
	friend int gmThing_set_pos( const gmCCArg &p);
	friend int gmThing_set_gravity( const gmCCArg &p);
	friend int gmThing_set_csys( const gmCCArg &p);
	friend int gmThing_set_velocity( const gmCCArg &p);

private:
	friend class ThingManager;
	ThingControlBlock *m_tcb;
};


// gmModelThing
// Description: Provides a "thing" which can also have a 3D model. All "things"
//	which need a basic 3D model or features provided by having a 3D model
//	may inherit from this class.
class gmModelThing : public gmThing
{
public:
	gmModelThing();
	virtual ~gmModelThing();

	void process( ThingManager *tm);

	bool reference_model( const std::string &key, ModelManager &m);
	void unreference_model();
	sk3DRomModel *model();

	void draw( gmRenderArgs *);

	std::string type() const { return std::string( "gmModelThing");}
	bool is_a( const std::string &s) const {
		if( s != gmModelThing::type()) return gmThing::is_a( s);
		return true;
	}

	int call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	//virtual void animate();
	void transform();
	void transform( const std::string &key) { transform( frame_id( key));}
	void transform( int fid);
	
	int frame_id( const std::string &key) const;

	sk3DRomDrawContext *draw_context() { return m_draw_context;}

	void animate();
	void animate( const int p_elapsed);
	int animate( const std::string &key, int fid, const int p_time,
		const int p_elapsed);

	void enable_animation( const bool b);

	int create_display_list( const std::string &key);
	GLint get_display_list( const std::string &key) const;
	int delete_display_list( const std::string &key);
	int clear_display_lists();
	int gen_display_list( const std::string &key);
	int draw_display_list( const std::string &key) const;


protected:
	ModelHandle *m_mh;
	sk3DRomDrawContext *m_draw_context;

	bool m_enable_animation;
	long long m_ani_clock_prev;

	std::map <const std::string, GLint> m_display_list;
	std::string m_use_display_list;

	static std::map <const std::string, int (*)( const gmCCArg &)> m_ccmd;
	friend int gmModelThing_set_model( const gmCCArg &p);
	friend int gmModelThing_transform( const gmCCArg &p);
	friend int gmModelThing_dump_matrix( const gmCCArg &p);
	friend int gmModelThing_load_identity( const gmCCArg &p);
	friend int gmModelThing_translate( const gmCCArg &p);
	friend int gmModelThing_rotate( const gmCCArg &p);
	friend int gmModelThing_scale( const gmCCArg &p);
	friend int gmModelThing_animate( const gmCCArg &p);
	friend int gmModelThing_set_dc_root_master( const gmCCArg &p);
	friend int gmModelThing_clear_display_lists( const gmCCArg &p);
	friend int gmModelThing_create_display_list( const gmCCArg &p);
	friend int gmModelThing_gen_display_list( const gmCCArg &p);
	friend int gmModelThing_use_display_list( const gmCCArg &p);
	friend int gmModelThing_use_vertex_array( const gmCCArg &p);
};


class gmPhyModelCOM
{
public:
	rgVector m_position;		// Real actual position

	// PSUEDO: m_position = bone_transformation( m_bound_position)
	rgVector m_bound_position;

	float m_mass;
	int m_bone_id;
	//rgVector velocity;
	//rgVector torque;

	gmPhyModelCOM() {
		m_mass = 1;
		m_bone_id = -1;
	}
};

// gmPhyModel
// Description: Similar to "gmModelThing" except that now physical properties
//	can be attributed to the model. For example, the "thing" can perform
//	collision detection with its model.
class gmPhyModel : public gmModelThing
{
public:
	gmPhyModel();
	virtual ~gmPhyModel();

	void process( ThingManager *tm);

	void draw( gmRenderArgs *);
	void draw_dynamic_bboxes();

	// Description: Applies gravity to the "things" center-of-mass (COM).
	//	COM is a point bound to a skeleton's root bone. A skeleton
	//	may have multiple root bones, however ONLY ONE bone can be
	//	used to define the center of mass. Inorder to allow gravity to
	//	affect all bones in the skeleton, all other bones must be
	//	either "bound" or "mounted" to the root bone containing the
	//	center of mass. m_draw_context->m_skout is the skeleton frame
	//	affected by gravity.
	void apply_gravity();

	// Description: Applies the "things" velocity to the center-of-mass
	//	(COM).
	void apply_velocity();

	// Description: Returns information about the models center-of-mass
	//	(COM). This includes things such as the current location of
	//	the COM relative to the world, and the mass of COM.
	gmPhyModelCOM &center_of_mass();

	// The center-of-mass (COM) defines the "things" position.
	rgCoordSystem cs() {
		m_cs.p = (center_of_mass()).m_position;
		return m_cs;
	}

	std::string type() const { return std::string( "gmPhyModel");}
	bool is_a( const std::string &s) const {
		if( s != gmPhyModel::type()) return gmModelThing::is_a( s);
		return true;
	}

	int call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	sk3DRomPhySkel *skel() { return m_skel;}

	// Description: Adds dynamic bounding boxes belonging to the given
	//	group number.
	bool add_dynamic_bboxes( const std::string &key);

	void clear_dynamic_bboxes();

	std::string dump_dyn_bboxes();

	int get_bbox( std::vector <rgVector> &v, int id);

	rgVector get_bbox_center( const int id);
	PhyBoundingPoint get_bbox_point( const int id);

	rgVector post_collision_vector( const rgVector v, const rgVector n)
		const;

	rgVector anti_intrusion( PhyCollisionFragment &f);

//	rgVector reflection( const rgVector v, const rgCollisionFragment &f)
//		const;

//	rgVector projection( const rgVector v, const rgCollisionFragment &f)
//		const;

protected:
	sk3DRomPhySkel *m_skel;
	gmPhyModelCOM m_com;

	PhyCollisionContext *m_cc;

	float m_bounciness;

	bool m_draw_dyn_bboxes;

	// m_dyn_bbox contains a list of bounding boxes that define what parts
	// of the model can collide with the enviornment or other models.
	// These bounding boxes are dynamic (meaning that skeletal
	// transformations affect the bounding boxes).
	// m_dyn_bbox (specifically) contains a list of rg3DRom brushes from
	// the model. When a bounding box is required, the index number is
	// looked up and the brush is converted into a bounding box.
	std::vector <int> m_dyn_bbox;

	static std::map <const std::string, int (*)( const gmCCArg &)> m_ccmd;
	friend int gmPhyModel_set_model( const gmCCArg &p);
	friend int gmPhyModel_load_identity( const gmCCArg &p);
	friend int gmPhyModel_scale( const gmCCArg &p);
	friend int gmPhyModel_set_center_of_mass( const gmCCArg &p);
	friend int gmPhyModel_add_dynamic_bboxes( const gmCCArg &p);
	friend int gmPhyModel_dump_dynamic_bboxes( const gmCCArg &p);
	friend int gmPhyModel_draw_dynamic_bboxes( const gmCCArg &p);
	friend int gmPhyModel_bounciness( const gmCCArg &p);

	friend class PhysicsEngine;
};


class phDebugger : public gmThing
{
public:
	void process( ThingManager *tm);
};

class gmDBGMessage : public gmThing
{
public:
	gmDBGMessage( char *msg) {
		strncpy( m_msg, msg, 159);
		m_msg[159] = '\0';
	}

	void process( ThingManager *tm) { std::cerr << m_msg;}

protected:
	char m_msg[160];
};

/*
// gmHologram
// Description: Provides a "thing" that has a displayable 3D model and no
//	"physical" existence. The "thing" cannot be hit, nor have gravity
//	or other similar physical effects applied to it. This type of "thing"
//	is usefull for non-interactive display models.
class gmHologram : public gmModelThing
{
public:
	
};
*/

class ThingControlBlock
{
public:
	gmThing *m_thing;			// The thing
	std::string m_key;

	// Process list thing is in
	std::list <ThingControlBlock *> *m_proc;
	std::list <ThingControlBlock *>::iterator m_iter;

	int m_prev;
	int m_next;

	ThingControlBlock() {
		m_thing = 0;
		m_proc = 0;
		m_prev = -1;
		m_next = -1;
		m_free = true;
		m_manager = 0;
	}

	bool is_free() const { return m_free;}

	ThingManager *manager() { return m_manager;}

private:
	friend class ThingManager;
	ThingManager *m_manager;
	bool m_free;
};

#define TM_MAX_PROC_LIST	17
#define TM_MAX_THINGS		16000
#define TM_ERR_GENERIC		-1
#define TM_ERR_SAME_KEY		-3
#define TM_ERR_NO_KEY		-4
#define TM_ERR_NO_THING		-5
#define TM_ERR_OPEN_FILE	-6
#define TM_ERR_WRITE_FILE	-7
class ThingManager
{
public:

	ThingManager( GameEngine *ge);
	~ThingManager();

	void clear();
	int create_thing( const std::string &key, gmThing *thg) {
		return create_thing( key, thg, -1);}
	int create_thing( const std::string &key, gmThing *thg, int p_pid);

	int remove_thing( const std::string &key);

	void process();
	GameEngine *ge() { return m_game_engine;}

	int count() const { return m_thing_count;}
	void dbg_stats( std::string &des) const;

	ThingControlBlock *get( int id) const;
	ThingControlBlock *get( const std::string &key) const;
	std::string dump_thing( const int id) const;
	std::string dump_thing( const std::string &key) const;
	int dump_all_things( std::string *s, const char *fname) const;

	int spawn( const std::string &key, const std::string &type);
	std::string list_spawnable() const;
	int add_to_spawn_list( const std::string &key, gmThing *(*fn)());

	int add_to_delete_list( const std::string &key, int (*fn)( gmThing *));

	void *get_module( const std::string &key) const;
	int load_module( const std::string &key, const std::string &fname,
		std::string &emsg);

	void move_thing_to_proc_list( ThingControlBlock *tcb, int pid);

	int index( ThingControlBlock *tcb) const {
		if( !tcb) return -1;
		return int( tcb - m_things);
	}

	int thing_list_check() const;

protected:
	void process_list( std::list <ThingControlBlock *> &p_proc);

	GameEngine *m_game_engine;

	int m_max_things;
	int m_thing_count;
	ThingControlBlock *m_things;
	std::list <ThingControlBlock *> m_proc[ TM_MAX_PROC_LIST];
	std::map <const std::string, int> m_thing_map;

	std::map <const std::string, gmThing *(*)()> m_spawn;
	std::map <const std::string, int (*)( gmThing *)> m_delete;

	std::map <const std::string, void *> m_module;

	int m_thing_head;
	int m_thing_free;

	void init_things();
	void move_thing_to_proc( ThingControlBlock &tcb,
		std::list <ThingControlBlock *> *p_proc);
	void remove_thing_from_proc( ThingControlBlock &tcb);
	void init_spawn_list();

	void html_dump_all_things( std::ostream &des) const;
	std::string html_dump_thing( const int id) const;
};

#endif
